# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
""" 
@ 模块     : 装饰器
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""


def singleton_pattern(cls):
    """ 
    [ 单例模式装饰器 ]

    ---
    描述:
        冠以该装饰器的类在整个程序的生命周期中只会被实例化一次, 存在实例时返回实例。

    """
    instances = {}
    
    def _singleton_pattern(*args, **kwargs):
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]
        
    return _singleton_pattern



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 名称常量
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""

class DirName: ...
class FileName: ...


class DirName:
    """
    [ 名称常量索引 :: 目录名称对象 ]
    
    """

    EngineeringData = 'EngineeringData'
    """ 工程数据目录 """

    EngineeringLogs = 'Logs'
    """ 工程日志目录 """

    EngineeringBaseCode = 'BaseCode'
    """ 工程底层编码目录 """


class FileName:
    """
    [ 名称常量索引 :: 文件名称对象 ]
    
    """

    EngineeringInit = 'AwakenEngineering.ini'
    """ 工程INI文件 """

    EngineeringConfig = 'AwakenConfig.yaml'
    """ 工程配置文件 """

    EngineeringDatabase = 'AwakenDatabase.db'
    """ 工程数据库文件 """

    ProjectInit = '_AwakenProject.ini'
    """ 项目INI文件 """



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 基础通用模块
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
import os

from .decorator import singleton_pattern

from ..common import AwakenDetails


class MessageRecorder: ...


@singleton_pattern
class MessageRecorder:
    """
    [ 错误记录器 ]

    ---
    描述:
        NULL

    """
    _error_warehouse: list
    """ 错误信息仓库 """

    def __init__(self) -> None:
        self._error_warehouse = []


    def record(self, message: str) -> None:
        """
        [ 记录 ]

        ---
        描述:
            NULL

        ---
        参数:
            message { str } : 记录信息文本

        """
        self._error_warehouse.append(message)


    def read(self) -> None:
        """
        [ 读取 ]

        ---
        描述:
            NULL

        """
        return self._error_warehouse


    def template_print(self) -> None:
        """
        [ 模板打印 ]

        ---
        描述:
            NULL

        """
        sym1 = '━'
        sym2 = '┃'
        sym3 = '┏'
        sym4 = '┗'
        sym5 = '┓'
        sym6 = '┛'
        tab_max_len = 50
        error_list = ['异常信息列表', ' ']
        error_number = 0

        # 处理初始错误信息
        for error in self._error_warehouse:
            error_list.append(f'error {error_number} ::')
            error = error.replace('\n', '')
            if len(error) > tab_max_len:
                while 1:
                    error_list.append(error[0: tab_max_len-3])
                    error = error[tab_max_len-3:]
                    error_list.append(error)
                    if len(error) >= tab_max_len:
                        continue
                    else:
                        error_list.append(error[tab_max_len-3:])
                        error_number += 1
                        break
            else:
                error_list.append(error)
                error_number += 1
            
            error_list.append(' ')

        # 消除空字符
        error_list = [error.replace(' ', '') for error in error_list if error != '']

        # 制表
        print(sym3 + sym1*(tab_max_len*2 + 3 - 20) + sym5)
        for error in error_list:
            offset = 0
            for s in error:
                if not u'\u4e00' <= s <= u'\u9fff':
                    offset += 1
            n = ' '*((tab_max_len*2 - len(error)*2 - 20) + offset)
            print(f'{sym2} {error} {n} {sym2}')
        print(sym4 + sym1*(tab_max_len*2 + 3 - 20) + sym6)


def window_template_output(message: list, name: str = None, clean: bool = True):
    """
    [ 窗口模板输出 ]
    
    ---
    描述:
        NULL

    ---
    参数:
        message { list } : 需要输出的文本列表
        name     { str } : 窗口模板名称
        clean   { bool } : 是否清空黑框

    """
    if clean:
        os.system('cls')

    max_limit = 50
    if not name:
        name = f'{AwakenDetails.Name} - {AwakenDetails.Version}'

    def calculate_offset(value):
        offset = 0
        for s in value:
            if u'\u4e00' <= s <= u'\u9fff':
                offset += 1
        return offset

    # 绘制窗口模板表格
    print('\n')
    print('┏' + '━'*max_limit + '┓')
    print('┃ ' + name + ' '*((max_limit - len(name) - 2) - calculate_offset(name)) + ' ┃')
    print('┣' + '━'*max_limit + '┫')
    print('┃' + ' '*max_limit + '┃')
    
    for mage in message:
        if not mage:
            print('┃' + ' '*max_limit + '┃')
        else:
            print('┃ ' + mage + ' '*((max_limit - len(mage) - 2) - calculate_offset(mage)) + ' ┃')

    print('┃' + ' '*max_limit + '┃')
    print('┗' + '━'*max_limit + '┛')


MESSAGE_RECORDER = MessageRecorder()
""" 错误记录器实例 """



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 基础设施通用模块
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
import os
import hashlib

from .const import CONST
from .message_recorder import window_template_output
from ..common import AwakenDetails
from ..kit.common import TASK_TEMPLATE_MAP


def environment_check_browser_driver_exists():
    """
    [ 环境检查::是否安装浏览器驱动 ]

    ---
    描述:
        NULL

    """
    dialog_box_title = '- 检查本地浏览器驱动 -'
    local_downloaded_browsers = []
    registered_browsers = [
        CONST.Type.Browser.Chromium,
        CONST.Type.Browser.Firefox,
        CONST.Type.Browser.Webkit,
    ]

    # ------------------------------------------------------------------------
    # 检查本地 playwright 环境是否安装浏览器驱动
    # 如果浏览器驱动不存在本地则下载
    # ------------------------------------------------------------------------
    window_template_output_text = [dialog_box_title, None]
    if CONST.Path.ENVIRONMENT_VARIABLE in os.environ:

        for dir_path in CONST.Path.PLAYWRIGHT_CACHE_PATH.glob('*'):
            local_downloaded_browsers.append(dir_path.stem.split('-')[0])

        for browser_name in registered_browsers:
            if browser_name not in local_downloaded_browsers:
                window_template_output_text.append(f'正在下载浏览器驱动 :: {browser_name}')
                window_template_output(window_template_output_text)
                os.system(f'playwright install { browser_name.lower() }')

    else:
        window_template_output_text.append(f'未能在系统环境变量中获取到用户配置文件路径 :: {CONST.Path.ENVIRONMENT_VARIABLE}')
        window_template_output(window_template_output_text)


def environment_check_dependent_files_exists():
    """
    [ 环境检查::是否创建工程依赖 ]

    ---
    描述:
        NULL
    
    """
    for path in [
        CONST.Path.DirPath.Data,
        CONST.Path.DirPath.Logs,
        CONST.Path.DirPath.BaseCode,
    ]:
        path.mkdir(parents=True, exist_ok=True)

    for path in [
        CONST.Path.FilePath.Config,
        CONST.Path.FilePath.Database,
    ]:
        path.touch(mode=0o777, exist_ok=True)


def create_project(name: str):
    """
    [ 创建项目 ]

    ---
    参数:
        name { str } : 项目名称

    ---
    返回:
        bool : 是否创建成功凭证
    
    """
    project_path = CONST.Path.CWD.joinpath(name)

    # ------------------------------------------------------------------------
    # 如果项目目录不存在则创建
    # ------------------------------------------------------------------------
    if not project_path.exists():
        project_path.mkdir(parents=True, exist_ok=True)
        project_path.joinpath(CONST.Name.FileName.ProjectInit).touch(mode=0o777, exist_ok=True)
        return True
    else:
        return False


def create_task(task_name: str, task_type: str):
    """ 
    [ 创建任务 ]

    ---
    参数:
        name { str } : 任务名称

    ---
    返回:
        bool : 是否创建成功凭证
    
    """
    task_type_upper = task_type.upper()
    awaken_file_name = f'{task_name}.awaken-{task_type}'
    awaken_file_path = CONST.Path.CWD.joinpath(awaken_file_name)

    # ------------------------------------------------------------------------
    # 如果任务不存在则创建
    # ------------------------------------------------------------------------
    if not awaken_file_path.exists():
        task_template = TASK_TEMPLATE_MAP[task_type_upper]
        task_template = task_template.replace('#TASK_NAME#', task_name)
        task_template = task_template.replace('#TASK_TYPE#', task_type_upper)
        awaken_file_path.touch(mode=0o777)
        awaken_file_path.write_bytes(bytes(task_template, encoding='UTF8'))
        return True
    else:
        return False


def encrypt_md5(string: str):
    """
    [ MD5加密 ]
    
    ---
    描述:
        NULL

    ---
    参数:
        string { str } : 需要加密的字符串

    """
    file_md5hash = hashlib.md5(string.encode('UTF-8'))
    return file_md5hash.hexdigest()


def window_template_output(message: list, name: str = None):
    """
    [ 窗口模板输出 ]
    
    ---
    描述:
        NULL

    ---
    参数:
        message { list } : 需要输出的文本列表
        name     { str } : 窗口模板名称

    """
    os.system('cls')  # 清空黑框

    max_limit = 50
    if not name:
        name = f'{AwakenDetails.Name} - {AwakenDetails.Version}'

    def calculate_offset(value):
        offset = 0
        for s in value:
            if u'\u4e00' <= s <= u'\u9fff':
                offset += 1
        return offset

    # 绘制窗口模板表格
    print('┏' + '━'*max_limit + '┓')
    print('┃ ' + name + ' '*((max_limit - len(name) - 2) - calculate_offset(name)) + ' ┃')
    print('┣' + '━'*max_limit + '┫')
    print('┃' + ' '*max_limit + '┃')
    
    for mage in message:
        if not mage:
            print('┃' + ' '*max_limit + '┃')
        else:
            print('┃ ' + mage + ' '*((max_limit - len(mage) - 2) - calculate_offset(mage)) + ' ┃')

    print('┃' + ' '*max_limit + '┃')
    print('┗' + '━'*max_limit + '┛')



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 通用字段关键字
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL
    
"""


class Config: ...


class Config:
    """
    [ 通用字段关键字索引 :: 配置 ]

    ---
    描述: 
        配置相关的字段。
        
    """

    Debug = 'debug'
    """ 调试模式 """

    IsWriteDatabase = 'is_write_database'
    """ 是否写入数据库 """

    EngineQueueMaxCount = 'engine_queue_max_count'
    """ 驱动队列阈值 """

    TaskQueueMaxCount = 'task_queue_max_count'
    """ 任务队列阈值 """

    CpuPropertyTime = 'cpu_property_time'
    """ CPU资源检查时间 """

    CpuPropertyCeiling = 'cpu_property_ceiling'
    """ CPU资源检查阈值 """

    LogLevelFloor = 'log_level_floor'
    """ 日志等级下限 """

    LogLevelDefault = 'log_level_default'
    """ 日志等级默认 """



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 接口常量
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""

class MessageCode: ...


class MessageCode:
    """
    [ 接口常量索引 :: 信息编码标识 ]

    ---
    描述: 
        定义 WebApi 返回的信息编码标识。

    """

    Success = 0
    """ 请求成功 """

    Unsuccess = -1
    """ 请求失败 """

    Error = -2
    """ 请求失败 """



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 命令行指令::创建任务
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
from ..baseic.const import CONST
from ..baseic.common import create_task
from ..baseic.message_recorder import window_template_output


def instruction_create_task(argv: list):
    """
    [ 命令行指令::创建任务 ]

    ---
    参数:
        argv { list } : 参数列表

    """
    dialog_box_title = '- 创建任务 -'

    if len(argv) >= 2:
        task_type: str = argv[0]
        if task_type.upper() == CONST.Type.Task.Web:
            ...
        else:
            window_template_output(
                [
                    dialog_box_title,
                    None,
                    '创建任务指令参数异常:',
                    '暂不支持的任务类型 !'
                ]
            )
            exit(0)

        success = []
        unsuccess = []
        for name in argv[1:]:
            result = create_task(name, task_type)
            if result:
                success.append(name)
            else:
                unsuccess.append(name)

            template = [
                dialog_box_title,
                None,
            ]
            if len(success) > 0:
                template.append('任务创建成功:')
                for name in success:
                    template.append(name)

            if len(unsuccess) > 0:
                if len(success) > 0:
                    template.append(None)
                template.append('任务创建失败:')
                for name in unsuccess:
                    template.append(name)

        window_template_output(template)

    else:
        window_template_output(
            [
                '- 创建任务 -',
                None,
                '创建任务指令参数异常:',
                None,
                '示例:',
                '>> awaken -task web task',
                '>> awaken -task web task1 task2',
                None,
                '目前支持的任务类型:',
                'web : WEB功能测试任务',
                'api : API功能测试任务',
            ]
        )
        exit(0)



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 脚本全局方法映射
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
import re

from .global_method import GlobalMethod


GLOBAL_METHOD_TYPE = type(GlobalMethod())
GLOBAL_METHOD_FUNCTION_MAP = {}


for fn, fv in GLOBAL_METHOD_TYPE.__dict__.items():
    if fn[0] != '_':
        fn = ''.join([n.title() for n in fn.rsplit('_')])
        cn_fn = re.findall(r'\[ (.*) \]', fv.__doc__)[0]
        GLOBAL_METHOD_FUNCTION_MAP.update({fn:fv})
        GLOBAL_METHOD_FUNCTION_MAP.update({cn_fn:fv})



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 命令行指令通用模块   
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
from ..baseic.const import CONST
from ..baseic.message_recorder import window_template_output


NUMBER_NESTING_LAYERS = 3
""" 工程目录嵌套最大层数 """


def confirm_root_directory_runtime_project():
    """
    [ 确认运行时工程的根目录 ]

    ---
    描述:
        NULL

    """
    cwd = CONST.Path.CWD

    for _ in range(NUMBER_NESTING_LAYERS):
        engineering_init = cwd.joinpath(CONST.Name.FileName.EngineeringInit)
        if engineering_init.exists():
            CONST.Path.CWD = cwd
            CONST.Path.DirPath.Data = cwd.joinpath(CONST.Name.DirName.EngineeringData)
            CONST.Path.DirPath.Logs = CONST.Path.DirPath.Data.joinpath(CONST.Name.DirName.EngineeringLogs)
            CONST.Path.DirPath.BaseCode = CONST.Path.DirPath.Data.joinpath(CONST.Name.DirName.EngineeringBaseCode)
            CONST.Path.FilePath.Init = cwd.joinpath(CONST.Name.FileName.EngineeringInit)
            CONST.Path.FilePath.Config = CONST.Path.DirPath.Data.joinpath(CONST.Name.FileName.EngineeringConfig)
            CONST.Path.FilePath.Database = CONST.Path.DirPath.Data.joinpath(CONST.Name.FileName.EngineeringDatabase)
            return True
        else:
            cwd = cwd.resolve().parent
            continue
    
    window_template_output(
        [
            '- 检查环境 -',
            None,
            '当前运行路径不是Awaken工程结构路径;',
            '请在工程根目录中执行工程初始化命令:',
            None,
            'awaken -init'
        ],
        clean=False
    )
    exit(0)



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : 自定义异常
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
import re

from ..baseic.message_recorder import MESSAGE_RECORDER
from ..baseic.message_recorder import window_template_output

class _AwakenBaseError(Exception):
    """ 
    [ 自定义异常基类 ]
    
    """

    def __init__(self, error_message: str):
        self.error = error_message
        self.err_name = re.findall(r'\[ (.*) \]', self.__doc__)[0]
        self.message = ''.join([self.err_name, ' :: ', str(error_message)])
        MESSAGE_RECORDER.record(self.message)

    def __str__(self) -> str:
        return self.message


class AwakenDependentFileIsMissing(_AwakenBaseError):
    """
    [ 依赖文件缺失 ]
    
    """
    def __str__(self) -> str:
        window_template_output(
            [
                f'- {self.err_name} -',
                None,
                '请先通过以下命令将当前目录初始化为工程 !',
                None,
                'awaken -init',
            ],
            clean=False,
        )
        return ''


class AwakenLogCreationError(_AwakenBaseError):
    """
    [ 日志创建异常 ]
    
    """


class AwakenLogOutputError(_AwakenBaseError):
    """
    [ 日志输出异常 ]
    
    """


class AwakenWebEngineError(_AwakenBaseError):
    """
    [ WEB引擎异常 ]
    
    """


class AwakenApiEngineError(_AwakenBaseError):
    """
    [ API引擎异常 ]
    
    """


class AwakenConvertCodelinError(_AwakenBaseError):
    """
    [ 转换代码行异常 ]
    
    """


class AwakenAnalysisBaseCodeError(_AwakenBaseError):
    """
    [ 解读底层编码异常 ]
    
    """


class AwakenTaskPretreatmentError(_AwakenBaseError):
    """
    [ 任务预处理异常 ]
    
    """

class AwakenWebApiServerError(_AwakenBaseError):
    """
    [ WebApi服务器异常 ]
    
    """


class AwakenDataBaseError(_AwakenBaseError):
    """
    [ 数据库异常 ]
    
    """


class AwakenWebEngineRunError(_AwakenBaseError):
    """
    [ WEB引擎运行异常 ]
    
    """




# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
@ 模块     : Awaken任务预处理程序
@ 作者     : chenjiancheng
@ 邮箱     : quinn.7@foxmail.com
@ 编写时间 : 2022-08-10

@ 模块描述 :
    NULL

"""
from .structural import AwakenTask
from .structural import AwakenCase
from .structural import AwakenCodeLine
from ...baseic.log import LOG
from ...baseic.const import CONST
from ...baseic.keyword import KEYWORD
from ...baseic.error import AwakenTaskPretreatmentError
from ...baseic.message_recorder import MESSAGE_RECORDER
from ...kit.common import ENGINE_TYPE_FUNCTION_MAP
from ...kit.global_method_map import GLOBAL_METHOD_FUNCTION_MAP


class TaskPreprocessor:
    """
    [ Awaken任务预处理程序 ]

    ---
    描述:
        NULL

    """
    
    _task: AwakenTask
    """ 任务对象 """

    def pretreatment(self, task: AwakenTask):
        self._task = task
        try:
            self._init_namespace_by_task_type()
            self._running_preprocessing_instruction()

        except BaseException:
            import traceback
            LOG.error(traceback.format_exc())
            MESSAGE_RECORDER.template_print()

        return self._task


    def _init_namespace_by_task_type(self):
        """
        [ 初始化任务命名空间 ]

        ---
        描述:
            NULL

        """
        self._task.test_cases: dict = {}
        self._task.namespace: dict = {
            KEYWORD.Script.Namespace.TaskType: None,
            KEYWORD.Script.Namespace.TaskName: None,
            KEYWORD.Script.Namespace.TaskDocs: None
        }

        if self._task.task_type == CONST.Type.Task.Web:
            self._task.namespace[KEYWORD.Script.Namespace.TaskType]    = CONST.Type.Task.Web
            self._task.namespace[KEYWORD.Script.Decorator.BrowserType] = CONST.Type.Browser.Chromium

        elif self._task.task_type == CONST.Type.Task.Api:
            self._task.namespace[KEYWORD.Script.Namespace.TaskType]    = CONST.Type.Task.Api


    def _running_preprocessing_instruction(self):
        """
        [ 运行预处理指令 ]

        ---
        描述:
            NULL

        """
        # --------------------------------------------------------------------
        # 解析全局&&引擎方法映射字典
        # --------------------------------------------------------------------
        self._task.global_function_map = GLOBAL_METHOD_FUNCTION_MAP
        self._task.engine_function_map = ENGINE_TYPE_FUNCTION_MAP[self._task.task_type]

        # --------------------------------------------------------------------
        # 循环全局域语句
        # 用例域语句将被存放至 test_cases 字典中等待引擎解析
        # --------------------------------------------------------------------
        for codeline in self._task.basecodes:
            awaken_codeline = AwakenCodeLine(codeline)

            # 公域语句
            if awaken_codeline.region == CONST.Interpreter.KEYWORD_IDENT_SCOPE_UNIVERSE:
                
                # 赋值逻辑
                if awaken_codeline.type == CONST.Interpreter.CodeLineType.Give:
                    self._give_global_assignment(awaken_codeline.give_name, awaken_codeline.give_value)

                # 执行并赋值逻辑
                elif awaken_codeline.type == CONST.Interpreter.CodeLineType.RGive:
                    result = self._running_common_function(awaken_codeline.funtion_name, awaken_codeline.funtion_value)
                    self._give_global_assignment(awaken_codeline.give_name, result)

                # 执行逻辑
                elif awaken_codeline.type == CONST.Interpreter.CodeLineType.Run:
                    self._running_common_function(awaken_codeline.funtion_name, awaken_codeline.funtion_value)

                # 声明逻辑
                elif awaken_codeline.type == CONST.Interpreter.CodeLineType.SCase:
                    awaken_case = AwakenCase(
                        awaken_codeline.number,
                        awaken_codeline.case_name,
                        awaken_codeline.case_docs,
                    )
                    self._task.test_cases[awaken_case.name] = awaken_case

            # 私域语句
            else:

                # 装饰器逻辑
                if awaken_codeline.type == CONST.Interpreter.CodeLineType.SDecorator:
                    self._task.test_cases[awaken_codeline.region].decorator.update({awaken_codeline.decorator_key : awaken_codeline.decorator_value})

                # 执行并赋值逻辑
                elif awaken_codeline.type == CONST.Interpreter.CodeLineType.RGive:
                    self.verify_validity_method(awaken_codeline)

                # 执行逻辑
                elif awaken_codeline.type == CONST.Interpreter.CodeLineType.Run:
                    self.verify_validity_method(awaken_codeline)

                # 其他语句
                else:
                    self._task.test_cases[awaken_codeline.region].steps.append(awaken_codeline)


    def _give_global_assignment(self, give_name: str, give_value: str):
        """
        [ 公域赋值封装 ]

        ---
        描述:
            NULL
        
        """
        give_path_node = give_name.split(CONST.Interpreter.GrammarSymbol.VariablePath)
        give_path_node_number = len(give_path_node)
        current_node = self._task.namespace

        if give_path_node_number > 1:
            i = 1
            for node in give_path_node:
                if i == give_path_node_number:
                    current_node.update({node: give_value})
                else:
                    if node not in current_node.keys():
                        current_node.update({node: {}})
                    current_node = current_node[node]
                    i += 1
        else:
            self._task.namespace.update({give_name : give_value})


    def _running_common_function(self, name: str, paras: list):
        """
        [ 公域运行方法封装 ]

        ---
        描述:
            NULL
        
        """
        try:
            if name not in self._task.global_function_map.keys():
                error_message = CONST.Error.Interpreter.GrammarMethodWrongful('#NAME#', name)
                raise AwakenTaskPretreatmentError(error_message)

            # ----------------------------------------------------------------
            # 解析方法参数
            # ----------------------------------------------------------------
            if len(paras) > 0:
                new_function_value = []

                for value in paras:
                    if isinstance(value, str):
                        value_quote_symbol_count = value[0:2].count(CONST.Interpreter.GrammarSymbol.Quote)
                        para = value[value_quote_symbol_count:]
                        para_statement_symbol_count = para.count(CONST.Interpreter.GrammarSymbol.VariablePath)

                        if para_statement_symbol_count != 0:
                            current_node = self._task.namespace
                            para_nodes = para.split(CONST.Interpreter.GrammarSymbol.VariablePath)

                            if len(para_nodes) > 1:
                                for node in para_nodes:
                                    try:
                                        current_node = current_node[node]
                                    except KeyError:
                                        error_message = CONST.Error.Interpreter.GrammarNamespaceNodeNotExist('#NODE#', node)
                                        AwakenTaskPretreatmentError(error_message)
                            else:
                                try:
                                    current_node = current_node[para_nodes[0]]
                                except:
                                    error_message = CONST.Error.Interpreter.GrammarNamespaceNodeNotExist('#NODE#', node)
                                    AwakenTaskPretreatmentError(error_message)

                            new_function_value.append(current_node)

                        else:
                            # 如果 para 存在于全局命名空间中则取出, 否则视为普通字符串
                            if para in self._task.namespace.keys():
                                new_function_value.append(self._task.namespace[para])
                                continue
                            else:
                                new_function_value.append(para)
                                continue

                    else:
                        new_function_value.append(value)

            result = self._task.global_function_map[name](self._task.global_method, *new_function_value)
            return result

        except KeyError:
            error_message = CONST.Error.Interpreter.GrammarMethodWrongful('#NAME#', name)
            raise AwakenTaskPretreatmentError(error_message)


    def verify_validity_method(self, awaken_codeline):
        """
        [ 校验方法的有效性 ]

        ---
        描述:
            NULL

        """
        function_map = None
        if awaken_codeline.funtion_region == CONST.Interpreter.CodeLineScopet.Global:
            function_map = self._task.global_function_map.keys()
        else:
            function_map = self._task.engine_function_map.keys()

        if awaken_codeline.funtion_name in function_map:
            self._task.test_cases[awaken_codeline.region].steps.append(awaken_codeline)
        else:
            error_message = CONST.Error.Interpreter.GrammarMethodWrongful.replace('#LINE#', awaken_codeline.number).replace('#NAME#', awaken_codeline.funtion_name)
            raise AwakenTaskPretreatmentError(error_message)
            